#ifndef __Weaver_h__
#define __Weaver_h__

#include "AutoThreadListener.h"
#include "Event.h"
#include "EventSets.h"
#include "ISocket.h"
#include "IWeaver.h"
#include "Map.h"
#include "ThreadStateGraph.h"
#include "UniqueVector.h"
#include "Vector.h"

class Weaver : public IWeaver
{
public:
    Weaver( ISelector &selector ) : _selector( selector )
    {
    }

    ~Weaver()
    {
        discardListeners();
    }

    void addThreadBlockingData( const String &threadName, Vector<Event> blockableEvents )
    {
        _threadsToBlockableEvents[threadName] = blockableEvents;
    }

    void addThreadStateGraph( const String &threadName, ThreadStateGraph &threadStateGraph )
    {
        _threadsToStateGraphs[threadName] = &threadStateGraph;

        States initialState;
        initialState.append( threadStateGraph._initialState );
        _threadsToLocationInGraph[threadName] = initialState;
    }

	void handleAcceptedSocket( ISocket &socket )
	{
        IThreadListener *threadListener = T::createThreadListener( *this, _selector, socket );
        storeThreadName( socket, threadListener );
        storeListener( threadListener );
    }

    void removeThread( IThreadListener *thread )
    {
        removeThreadFromDataStructures( thread );
        destroyThread( thread );

        if ( everyoneSynchornized() )
        {
            selectNextEvent();
        }
        else
        {
            try
            {
                attemptEagerSynchronization();
            }
            catch ( const EagerSynchronizationError &e )
            {
                clearDataStructures();
            }
        }
    }

    void bSync( IThreadListener *thread,
                const Vector<Event> &requested,
                const Vector<Event> &watched,
                const Vector<Event> &blocked )
    {
        storeEvents( thread, requested, watched, blocked );

        if ( everyoneSynchornized() )
        {
            selectNextEvent();
        }
        else
        {
            try
            {
                attemptEagerSynchronization();
            }
            catch ( const EagerSynchronizationError &e )
            {
                clearDataStructures();
            }
        }
    }

private:
    class EagerSynchronizationError
    {
    };

    typedef Vector<Event> Events;
    typedef Vector<String> States;

    ISelector &_selector;
    Vector<IThreadListener *> _threadListeners;
    Map<IThreadListener *, EventSets> _threadsToEvents;
    UniqueVector<Event> _requested;
    UniqueVector<Event> _blocked;
    unsigned _selectedEvent;
    Map<String, Events> _threadsToBlockableEvents;
    Map<String, ThreadStateGraph *> _threadsToStateGraphs;
    Map<String, States> _threadsToLocationInGraph;
    Map<IThreadListener *, String> _threadsToNames;
    Vector<IThreadListener *> _unsynchronizedThreads;
    Vector<Event> _candidates;

    void storeListener( IThreadListener *threadListener )
    {
        _threadListeners.append( threadListener );
    }

    void storeThreadName( ISocket &socket, IThreadListener *threadListener )
    {
        HeapData nameData;
        socket.receive( nameData );
        _threadsToNames[threadListener] = String( (char *)nameData.data(), nameData.size() );
    }

    bool everyoneSynchornized() const
    {
        return _threadsToEvents.size() == _threadListeners.size();
    }

    void attemptEagerSynchronization()
    {
        collectRequested();
        collectBlocked();
        collectCandidates();

        considerUnsynchronziedThreads();

        if ( existCandidates() )
        {
            selectFirstCandidate();
            announceEvent();
        }

        clearDataStructures();
    }

    bool existCandidates() const
    {
        return !_candidates.empty();
    }

    void selectFirstCandidate()
    {
        _selectedEvent = _candidates[0].type();
    }

    void considerUnsynchronziedThreads()
    {
        findUnsynchronizedThreads();
        for ( unsigned i = 0; i < _unsynchronizedThreads.size(); ++i )
        {
            considerUnsynchronziedThread( _unsynchronizedThreads[i] );
        }
    }

    void considerUnsynchronziedThread( IThreadListener *thread )
    {
        String threadName = _threadsToNames[thread];

        if ( _threadsToBlockableEvents.exists( threadName ) )
        {
            checkBlockableEvents( threadName );
        }
        else if ( _threadsToStateGraphs.exists( threadName ) )
        {
            checkStateGraph( threadName );
        }
        else
        {
            throw EagerSynchronizationError();
        }
    }

    void checkBlockableEvents( String &threadName )
    {
        Events blockableEvents = _threadsToBlockableEvents[threadName];

        for ( unsigned i = 0; i < blockableEvents.size(); ++i )
        {
            if ( _candidates.exists( blockableEvents[i] ) )
            {
                _candidates.erase( blockableEvents[i] );
            }
        }
    }

    void checkStateGraph( String &threadName )
    {
        States currentStateLabels = _threadsToLocationInGraph[threadName];
        Events blockedEventsApproximation = approximateBlockedEvents( threadName, currentStateLabels );

        for ( unsigned i = 0; i < blockedEventsApproximation.size(); ++i )
        {
            if ( _candidates.exists( blockedEventsApproximation[i] ) )
            {
                _candidates.erase( blockedEventsApproximation[i] );
            }
        }
    }

    Events approximateBlockedEvents( String &threadName, States &possibleStates )
    {
        Events approximation = _threadsToStateGraphs[threadName]->_graph[possibleStates[0]]._blockedEvents;

        for ( unsigned i = 1; i < possibleStates.size(); ++i )
        {
            Events blockedEvents = _threadsToStateGraphs[threadName]->_graph[possibleStates[i]]._blockedEvents;
            Events copy = approximation;

            for ( unsigned j = 0; j < copy.size(); ++j )
            {
                if ( !blockedEvents.exists( copy[j] ) )
                {
                    approximation.erase( copy[j] );
                }
            }
        }

        return approximation;
    }

    void findUnsynchronizedThreads()
    {
        for ( unsigned i = 0; i < _threadListeners.size(); ++i )
        {
            if ( !_threadsToEvents.exists( _threadListeners[i] ) )
            {
                _unsynchronizedThreads.append( _threadListeners[i] );
            }
        }
    }

    void storeEvents( IThreadListener *thread,
                      const Vector<Event> &requested,
                      const Vector<Event> &watched,
                      const Vector<Event> &blocked )
    {
        _threadsToEvents[thread] = EventSets( requested, watched, blocked );
    }

    void discardListeners()
    {
        for ( unsigned i = 0; i < _threadListeners.size(); ++i )
        {
            T::discardThreadListener( _threadListeners[i] );
        }
    }

    void collectRequested()
    {
        Map<IThreadListener *, EventSets>::iterator it;

        for ( it = _threadsToEvents.begin(); it != _threadsToEvents.end(); ++it )
        {
            _requested += it->second._requested;
        }
    }

    void collectBlocked()
    {
        Map<IThreadListener *, EventSets>::iterator it;

        for ( it = _threadsToEvents.begin(); it != _threadsToEvents.end(); ++it )
        {
            _blocked += it->second._blocked;
        }
    }

    void selectEvent()
    {
        collectRequested();
        collectBlocked();
        findRequestedNotBlocked();
    }

    void findRequestedNotBlocked()
    {
        for ( unsigned i = 0; i < _requested.size(); ++i )
        {
            if ( enabledEvent( _requested[i] ) )
            {
                _selectedEvent = _requested[i].type();
                return;
            }
        }

        throw Error( Error::SELECTION_DEADLOCKED );
    }

    void collectCandidates()
    {
        for ( unsigned i = 0; i < _requested.size(); ++i )
        {
            if ( enabledEvent( _requested[i] ) )
            {
                _candidates.append( _requested[i] );
            }
        }
    }

    bool enabledEvent( Event event )
    {
        return !_blocked.exists( event );
    }

    bool threadRequestedOrWatched( IThreadListener *thread )
    {
        return threadRequested( thread ) || threadWatched( thread );
    }

    bool threadRequested( IThreadListener *thread )
    {
        for ( unsigned i = 0; i < _threadsToEvents[thread]._requested.size(); ++i )
        {
            if ( _threadsToEvents[thread]._requested[i] == _selectedEvent )
            {
                return true;
            }
        }

        return false;
    }

    bool threadWatched( IThreadListener *thread )
    {
        for ( unsigned i = 0; i < _threadsToEvents[thread]._watched.size(); ++i )
        {
            if ( _threadsToEvents[thread]._watched[i] == _selectedEvent )
            {
                return true;
            }
        }

        return false;
    }

    void announceEvent()
    {
        Event event( _selectedEvent );

        announceToSynchronizedThreads( event );
        announceToUnsynchronizedThreads( event );
        updateStateGraphs( event );
    }

    void announceToSynchronizedThreads( Event &event )
    {
        Map<IThreadListener *, EventSets>::iterator it;
        Map<IThreadListener *, EventSets> copy = _threadsToEvents;

        for ( it = copy.begin(); it != copy.end(); ++it )
        {
            if ( threadRequestedOrWatched( it->first ) )
            {
                it->first->announceChosenEvent( event );
                _threadsToEvents.erase( it->first );
            }
        }
    }

    void announceToUnsynchronizedThreads( Event &event )
    {
        for ( unsigned i = 0; i < _unsynchronizedThreads.size(); ++i )
        {
            _unsynchronizedThreads[i]->announceChosenEvent( event );
        }
    }

    void clearDataStructures()
    {
        _requested.clear();
        _blocked.clear();
        _unsynchronizedThreads.clear();
        _candidates.clear();
    }

    void removeThreadFromDataStructures( IThreadListener *thread )
    {
        if ( _threadListeners.exists( thread ) )
        {
            _threadListeners.erase( thread );
        }
        if ( _threadsToEvents.exists( thread ) )
        {
            _threadsToEvents.erase( thread );
        }
        if ( _threadsToNames.exists( thread ) )
        {
            _threadsToNames.erase( thread );
        }
    }

    void destroyThread( IThreadListener *thread )
    {
        T::discardThreadListener( thread );
    }

    void selectNextEvent()
    {
        selectEvent();
        announceEvent();
        clearDataStructures();
    }

    void updateStateGraphs( Event &event )
    {
        Map<String, ThreadStateGraph *>::iterator it;

        for ( it = _threadsToStateGraphs.begin(); it != _threadsToStateGraphs.end(); ++it )
        {
            updateStateGraph( it->first, it->second, event );
        }
    }

    void updateStateGraph( String threadName, ThreadStateGraph *threadStateGraph, Event &event )
    {
        States currentStateLabels = _threadsToLocationInGraph[threadName];

        States newStateLabels;
        for ( unsigned i = 0; i < currentStateLabels.size(); ++i )
        {
            ThreadState *state = &threadStateGraph->_graph[currentStateLabels[i]];
            // state->dump();

            if ( state->_requestedEvents.exists( event ) || state->_watchedEvents.exists( event ) )
            {
                States successorLabels = state->labelOfSuccessors( event );
                newStateLabels += successorLabels;
            }
            else
            {
                newStateLabels.append( currentStateLabels[i] );
            }
        }

        _threadsToLocationInGraph[threadName] = newStateLabels;
    }
};

#endif // __Weaver_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
