#include <cxxtest/TestSuite.h>

#include "BProgram.h"
#include "Event.h"
#include "MockBThread.h"
#include "MockEngineFactory.h"
#include "MockErrno.h"

class MockForBProgram :
    public MockErrno,
    public MockEngineFactory
{
public:
};

class BProgramTestSuite : public CxxTest::TestSuite
{
public:
	MockForBProgram *mock;
    MockEngine *engine;

	void setUp()
	{
		TS_ASSERT( mock = new MockForBProgram );

        engine = &( mock->mockEngine );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_run_program()
    {
        BProgram *bProgram;

        TS_ASSERT( bProgram = new BProgram );

        MockBThread thread1;
        MockBThread thread2;
        MockBThread thread3;

        TS_ASSERT_THROWS_NOTHING( bProgram->addThread( thread1 ) );
        TS_ASSERT_THROWS_NOTHING( bProgram->addThread( thread2 ) );
        TS_ASSERT_THROWS_NOTHING( bProgram->addThread( thread3 ) );

        TS_ASSERT_THROWS_NOTHING( bProgram->runProgram() );

        TS_ASSERT( !engine->runWasCalled );

        TS_ASSERT( thread1.runWasCalled );
        TS_ASSERT( thread2.runWasCalled );
        TS_ASSERT( thread3.runWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete bProgram );

        TS_ASSERT( engine->wasDiscarded );
    }

    void test_run_engine()
    {
        BProgram *bProgram;

        TS_ASSERT( bProgram = new BProgram );

        TS_ASSERT( engine->wasCreated );

        TS_ASSERT_THROWS_NOTHING( bProgram->runEngine() );

        TS_ASSERT( engine->runWasCalled );

        TS_ASSERT_THROWS_NOTHING( delete bProgram );

        TS_ASSERT( engine->wasDiscarded );
    }

    void test_add_blocking_data()
    {
        BProgram bProgram;

        Vector<Event> events;
        events.append( Event( 5 ) );
        events.append( Event( 6 ) );

        TS_ASSERT_THROWS_NOTHING( bProgram.addThreadBlockingData( "Jefferson", events ) );

        TS_ASSERT_EQUALS( engine->lastThreadName, "Jefferson" );
        TS_ASSERT_EQUALS( engine->lastBlockableEvents.size(), 2U );
        TS_ASSERT_EQUALS( engine->lastBlockableEvents[0].type(), 5U );
        TS_ASSERT_EQUALS( engine->lastBlockableEvents[1].type(), 6U);
    }

    void test_add_blocking_data__state_graph()
    {
        BProgram bProgram;

        ThreadStateGraph threadStateGraph;

        threadStateGraph._events.append( Event( 1 ) );
        threadStateGraph._events.append( Event( 2 ) );

        threadStateGraph._initialState = "a";

        ThreadState a;
        ThreadState b;

        a._label = "a";
        a._requestedEvents.append( Event( 1 ) );
        a._blockedEvents.append( Event( 2 ) );
        a._successors.append( Transition( Event( 1 ), "b" ) );

        b._label = "b";
        b._watchedEvents.append( Event( 2 ) );
        b._successors.append( Transition( Event( 2 ), "b" ) );
        b._predecessors.append( Transition( Event( 1 ), "a" ) );
        b._history.append( Event( 1 ) );

        threadStateGraph._graph["a"] = a;
        threadStateGraph._graph["b"] = b;

        TS_ASSERT_THROWS_NOTHING( bProgram.addThreadStateGraph( "Jefferson", threadStateGraph ) );

        TS_ASSERT_EQUALS( engine->lastThreadName, "Jefferson" );
        TS_ASSERT_EQUALS( engine->lastThreadStateGraph, &threadStateGraph );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
